/*
 *  solutions.h
 *  calcul des solutions d'un polynôme sur Q
 */

#include <stdlib.h>
#include "fractions.h"
#include "liste.h"

// ci-dessous, l'adaptation des fonctions dans liste.h au fractions
typedef struct fnode {
  frac x;
  struct fnode *next;
} fnode;

fnode* fprepend(fnode *head, frac i) {
  fnode *n;
  n = malloc(sizeof(fnode));
  n->x = i;
  n->next = head;
  return n;
}

fnode* fappend(fnode *head, frac i) {
  if (head == NULL) {
    head = malloc(sizeof(fnode));
    head->x = i;
    head->next = NULL;
    return head;
  }
  fnode *n = head;
  while(n->next != NULL) n = n->next;
  n->next = malloc(sizeof(fnode));
  n = n->next;
  n->x = i;
  n->next = NULL;
  return head;
}

void affichelistef(fnode *head) {
  if (head == NULL) {
    printf("< >");
    return;
  }
  printf ("< ");
  affichef(head->x);
  head = head->next;
  while (head != NULL) {
    printf(", ");
    affichef(head->x);
    head = head->next;
  }
  printf(" >");
}

fnode* frmfirst(fnode *head) {
  if (head == NULL) return NULL;
  fnode *n = head->next;
  free(head);
  return n;
}


fnode* frmlast(fnode *head) {
  if (head == NULL) return NULL;
  if (head->next == NULL) {
    free(head);
    return NULL;
  }
  fnode *n = head;
  while (n->next->next != NULL) n = n->next;
  free(n->next);
  n->next = NULL;
  return head;
}

// lecture d'un polynôme comme liste de coefficients
// l'ordre sera a_0 + ··· + a_n X^n, ainsi que
// l'evaluation peut être fait sans garder ou calculer le degrée
// (voir la fonction eval)
// donc en utilisant prepend la lecture est en envers, de a_n à a_0
// On impose que a_n != 0. Pour le polynôme 0, on a une liste vide.
// Les coefficients sont simplifiés
fnode* entrepol() {
  unsigned int n;
  printf("svp, tapez le degrée du polynôme : ");
  scanf("%u",&n); // %u : chaîne de format pour les unsigned
                  //      (pas grave si on met %d)
  unsigned int i;
  fnode *res = NULL;
  frac f;
  printf("coefficient de x^%u : ",n);
  liref(&f);
  if (n>0) {  
      while (egalzerof(f)) {
      printf("le coefficient principal doit être non nul, svp retapez-le : ");
      liref(&f);
    }
  } else {  // si n == 0, on permet l'entrée de zero pour pouvoir
            // lire le polynôme nul
    if (egalzerof(f)) return NULL;
  }
  simplf(&f); // simplification
  res = fappend(res,f);
  for (i = n-1; i != -1; i--) { // rappel : si i est un unsigned int
              // i >= 0 est toujours vrai
    printf("coefficient de x^%u : ",i);
    liref(&f);
    simplf(&f); // simplification
    res = fprepend(res,f);
  }
  return res;
}

// evaluation d'un polynôme en une fraction x
frac eval(fnode *poly,frac x) {
  unsigned int i=0; // degrée du monôme parcourru
  frac res;
  res.num = 0;  // initialisation à 0/1 de res
  res.den = 1;
  while (poly != NULL) {
    res = sommef(res, prodf(poly->x,puissf(x,i)) ); // res=res + a_i*x^i
    poly = poly->next;
    i++;
  }
  return res;
}

unsigned int racine(fnode *poly, frac x) {
  return egalzerof(eval(poly,x));
}

// les fonction suivantes sont fonctions d'affichage des polynômes
// il faut toujours mieux que le perfectionnement de ces fonctions
// soit fait à la fin, quand le programme déjà fonctionne...

// affichage d'un monôme
// on affiche le coefficient si c'est different de 1 ou -1
// on affiche X seulement si l'exponant est positif, et cela est
// affiché seulement si plus grand que 1
void affichemon(frac coeff,unsigned int i) {
  if (i == 0) {
    affichef(coeff);
    return;
  }
  if(!egalunf(coeff)) {
    if(!egalmoinsunf(coeff)) {
      affichef(coeff);
    } else {
      printf("-");
    }
  }
  if(i > 1) {
    printf("X^%d",i);
  } else {
    printf("X");
  }
}

// affichage d'un polynôme
// on saute les coefficients nuls
// pour les monômes suivants le premier affiché, on écrit "+" si
// le coefficients est positif (sinon le "-" est deja écrit par
// l'affichage du coefficient
void affichepol(fnode *head) {
  if (head == NULL) { // polynôme nul
    printf("0");
    return;
  }
  unsigned int i = 0; // exponant du monôme
  while (egalzerof(head->x)) {  // on saute les premiers coefficients
                                // nuls.
                                // head ne peut pas être nul
                                // car il y a au moins un coefficient
                                // non nul dans le polynôme
                                // (il faut y penser !)
    head = head->next;
    i++;
  }
  affichemon(head->x,i); // affichage du premier monôme non nul
  head = head->next;
  i++;
  while (head != NULL) { // cycle sur le reste des monômes
    if (!egalzerof(head->x)) { // on saute les monômes nuls
      if (pgzerof(head->x)) printf(" +");
      else printf(" ");
      affichemon(head->x,i);
    }
    head = head->next;
    i++;
  }
}

// creation d'une liste des diviseurs de n, en assumant n != 0
node* diviseurs(int n) {
  node *res = NULL;
  if (n < 0) n = -n;  // n <- |n|
  res = prepend(res,1);   // 1 est toujours un diviseur de n
  if (n == 1) return res; // si n = ±1, alors 1 est le seul diviseur
  unsigned int i;
  for (i = 2; i <= n/2; i++) {
    if (n%i == 0) res = prepend(res,i);
  }
  res = prepend(res,n); // n est toujours un diviseur de n
  return res;
}

/*
 * fonction qui crée les deux generateurs des candidats à racine.
 * On assume que @poly est un polynôme a_0 + ··· + a_n X^n avec
 * n > 0, a_0 != 0 et a_n != 0
 * name : generateurs
 * @poly :  polynôme avec les proprietés ci-dessu (en particulier
 *          poly != NULL)
 * @res1 :  pointeur à un entier où on va mettre a_0·m, où m est
 *          le plus petit commun facteur des coefficients
 * @res2 :  pointeur à un entier où on va mettre a_n·m
 */
void generateurs(fnode *poly, int *res1, int *res2) {
  unsigned int ppcf = 1; // 1 est element neutre de ppcm
  fnode *n = poly;
  while (n->next != NULL) {
    ppcf = ppcm(ppcf, n->x.den);
    n = n->next;
  }
  ppcf = ppcm(ppcf, n->x.den);
  *res1 = ppcf/poly->x.den * poly->x.num;
  *res2 = ppcf/n->x.den * n->x.num;
}
// la fonction ci-dessu est un exemple de une fonction qui moralement
// renvoie deux resultats. On les fait en utilisant des pointeurs


// fonction qui renvoie une liste de toutes les combinaisons ±p/q
// où p|a et q|b
fnode *candidats(int a, int b) {
  node *diva = diviseurs(a);
  node *divb = diviseurs(b);
  frac f;
  fnode *res = NULL; // initialisation à la liste vide
  node *app = divb;   // on sauvegarde le debut de la liste divb
                      // à reinitialiser à chaque tour
  while (diva != NULL) {
    while (divb != NULL) {
      f.num = diva->x;
      f.den = divb->x;
      if (pgcd(f.num,f.den) == 1) { // on sauvegarde
                                    // seulement si la fraction
                                    // est déjà simplifié
        res = fprepend(res, f); // candidat positif...
        f.num = -f.num;
        res = fprepend(res, f); // et candidat negatif
      }
      divb = divb->next;
    }
    divb = app;           // à chaque tour de divb, on le reinitialise
    diva = rmfirst(diva); // et jette l'element de diva, car
                          // on n'y reviendra plus
  }
  while (divb != NULL) divb = rmfirst(divb);  // à la fine on jette
                                              // aussi toute divb
  return res;
}

// fonciton qui renvoie les solutions rationelles d'un polynôme non nul
fnode *solutions(fnode* poly) {
  fnode *res = NULL;
  if (egalzerof(poly->x)) { // ici on assume que poly est non nul
                            // à controler avant l'appel de cette
                            // fonction
                            // si a_0 == 0 alors 0 est solutions
                            // et les autres solutions sont les mêmes
                            // de a_i + ··· + a_n X^(n-i)
                            // où a_i est le premier coefficient non nul
    frac zero;
    zero.num = 0;
    zero.den = 1;
    res = fprepend(res, zero); // 0 est solution
    do {
      poly = poly->next;
    } while (egalzerof(poly->x)); // on cherche le premier a_i non nul
                          // poly devienne le polynôme a_i+···+a_n X^(n-i)
  }
  if (poly->next == NULL) {
    return res; // or, si poly est un polynôme constant (mais non nul,
                // car on l'a déjà contrôlè), alors on a fini
                // (res est vide ou <0>)
  }
  // maintenant on est dans les hypothèses sur poly pour utiliser
  // la fonction generateurs et le reste
  int a,b;
  generateurs(poly,&a,&b);
  fnode *cand = candidats(a,b);
  fnode *n;
  // on deplace les racines du cand à res
  // plutôt que utiliser frmfirst et prepend (qui detruiraient et
  // créeraient une structure fnode inutilement), si on a une racine
  // on deplace la structure fnode
  while(cand != NULL) {
    if (racine(poly, cand->x)) {
      n = cand->next;   // on sauvegard le deuxième élement de cand
      cand->next = res;
      res = cand;       // maintenant res à cand comme premier élement
      cand = n;         // et cand le perd
      // essayez à voir les operations ci-dessu à l'aide 
      // de la visualisation graphique de la mémoire
    } else {
      cand = frmfirst(cand);  // si le candidat n'est pas une racine,
                              // on le jette
    }
  }
  return res;
}
